# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utilities for validation of user profiles and preferences."""

from .dicts import UserPreferenceDict, UserProfileDict, ValidatedDict
from .schemas import UserPreferencesSchema, UserProfileSchema

__all__ = (
    "UserPreferenceDict",
    "UserPreferencesSchema",
    "UserProfileDict",
    "UserProfileSchema",
    "ValidatedDict",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""E2E/integration tests for Invenio-Accounts."""

from time import sleep
from urllib.request import urlopen

import flask

from invenio_accounts import testutils


def test_live_server(live_server):
    """Test that the login page of `live_server` is reachable.

    The `live_server` fixture is provided by pytest-flask.  When a test using
    said fixture is run, the return object of the `app` fixture is run in a
    background process.  For this test pytest will use the `app` fixture
    defined in `./conftest.py` instead of the one in `../conftest.py`.
    """
    # With pytest-flask we don't need to be in the application context
    # to use `flask.url_for`.
    url = flask.url_for('security.login', _external=True)
    response = urlopen(url)
    assert response
    assert response.code == 200


def test_webdriver_not_authenticated(live_server, env_browser):
    """Test unauthenticated webdriver client redirect.

    Test that an unauthenticated webdriver client is redirected from
    the change password page to the login page.
    """
    browser = env_browser
    browser.get(flask.url_for('security.change_password', _external=True))
    assert (flask.url_for('security.login', _external=True) in
            browser.current_url)


def test_user_registration(live_server, env_browser):
    """E2E user registration and login test."""
    browser = env_browser
    # 1. Go to user registration page
    browser.get(flask.url_for('security.register', _external=True))
    assert (flask.url_for('security.register', _external=True) in
            browser.current_url)
    # 2. Input user data
    signup_form = browser.find_element_by_name('register_user_form')
    input_email = signup_form.find_element_by_name('email')
    input_password = signup_form.find_element_by_name('password')
    # input w/ name "email"
    # input w/ name "password"
    user_email = 'test@example.org'
    user_password = '12345_SIx'
    input_email.send_keys(user_email)
    input_password.send_keys(user_password)

    # 3. submit form
    signup_form.submit()
    sleep(1)  # we need to wait after each form submission for redirect

    # ...and get redirected to the "home page" ('/')
    # This isn't a very important part of the process, and the '/' url isn't
    # even registered for the Invenio-Accounts e2e app. So we don't check it.

    # 3.5: After registering we should be logged in.
    browser.get(flask.url_for('security.change_password', _external=True))
    assert (flask.url_for('security.change_password', _external=True) in
            browser.current_url)

    # 3.5: logout.
    browser.get(flask.url_for('security.logout', _external=True))
    assert not testutils.webdriver_authenticated(browser)

    # 4. go to login-form
    browser.get(flask.url_for('security.login', _external=True))
    assert (flask.url_for('security.login', _external=True) in
            browser.current_url)
    login_form = browser.find_element_by_name('login_user_form')
    # 5. input registered info
    login_form.find_element_by_name('email').send_keys(user_email)
    login_form.find_element_by_name('password').send_keys(user_password)
    # 6. Submit!
    # check if authenticated at `flask.url_for('security.change_password')`
    login_form.submit()
    sleep(1)

    assert testutils.webdriver_authenticated(browser)

    browser.get(flask.url_for('security.change_password', _external=True))
    assert (flask.url_for('security.change_password', _external=True) in
            browser.current_url)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test account views."""

import base64
from unittest import mock

from flask import url_for
from flask_babelex import gettext as _
from flask_login import COOKIE_NAME, LoginManager
from flask_security import url_for_security
from flask_security.forms import LoginForm
from flask_security.views import _security

from invenio_accounts.models import SessionActivity
from invenio_accounts.testutils import create_test_user


def test_no_log_in_message_for_logged_in_users(app):
    """Test the password reset form for logged in users.

    Password reset form should not show log in or sign up messages for logged
    in users.
    """
    with app.app_context():
        forgot_password_url = url_for_security('forgot_password')

    with app.test_client() as client:
        log_in_message = _('Log in').encode('utf-8')
        sign_up_message = _('Sign up').encode('utf-8')

        resp = client.get(forgot_password_url)
        assert resp.status_code == 200
        assert log_in_message in resp.data
        assert sign_up_message in resp.data

        test_email = 'info@inveniosoftware.org'
        test_password = 'test1234'
        resp = client.post(url_for_security('register'), data=dict(
            email=test_email,
            password=test_password,
        ), environ_base={'REMOTE_ADDR': '127.0.0.1'})

        resp = client.post(url_for_security('login'), data=dict(
            email=test_email,
            password=test_password,
        ))

        resp = client.get(forgot_password_url, follow_redirects=True)
        if resp.status_code == 200:
            # This behavior will be phased out in future Flask-Security
            # release as per mattupstate/flask-security#291
            assert log_in_message not in resp.data
            assert sign_up_message not in resp.data
        else:
            # Future Flask-Security will redirect to post login view when
            # authenticated user requests password reset page.
            assert resp.data == client.get(
                app.config['SECURITY_POST_LOGIN_VIEW']).data


def test_view_list_sessions(app, app_i18n):
    """Test view list sessions."""
    with app.test_request_context():
        user1 = create_test_user(email='user1@invenio-software.org')
        user2 = create_test_user(email='user2@invenio-software.org')

        with app.test_client() as client:
            client.post(url_for_security('login'), data=dict(
                email=user1.email,
                password=user1.password_plaintext,
            ))

        with app.test_client() as client:
            client.post(url_for_security('login'), data=dict(
                email=user2.email,
                password=user2.password_plaintext,
            ))

            # get the list of user 2 sessions
            url = url_for('invenio_accounts.security')
            res = client.get(url)
            assert res.status_code == 200

            # check session for user 1 is not in the list
            sessions_1 = SessionActivity.query.filter_by(
                user_id=user1.id).all()
            assert len(sessions_1) == 1
            assert sessions_1[0].sid_s not in res.data.decode('utf-8')

            # check session for user 2 is in the list
            sessions_2 = SessionActivity.query.filter_by(
                user_id=user2.id).all()
            assert len(sessions_2) == 1
            assert sessions_2[0].sid_s in res.data.decode('utf-8')

            # test user 2 to delete user 1 session
            url = url_for('invenio_accounts.revoke_session')
            res = client.post(url, data={'sid_s': sessions_1[0].sid_s})
            assert res.status_code == 302
            assert SessionActivity.query.filter_by(
                user_id=user1.id, sid_s=sessions_1[0].sid_s).count() == 1

            # test user 2 to delete user 1 session
            url = url_for('invenio_accounts.revoke_session')
            res = client.post(url, data={'sid_s': sessions_2[0].sid_s})
            assert res.status_code == 302
            assert SessionActivity.query.filter_by(
                user_id=user1.id, sid_s=sessions_2[0].sid_s).count() == 0


def test_login_remember_me_disabled(app, users):
    """Test login remember me is disabled."""
    email = users[0]['email']
    password = users[0]['password']
    _security.login_form = LoginForm
    with app.test_client() as client:
        res = client.post(
            url_for_security('login'),
            data={'email': email, 'password': password, 'remember': True},
            environ_base={'REMOTE_ADDR': '127.0.0.1'})
        # check the remember_me cookie is not there
        name = '{0}='.format(COOKIE_NAME)
        assert all([
            name not in val for val in res.headers.values()])
        # check the session cookie is still there
        assert any([
            'session=' in val for val in res.headers.values()])


def test_login_from_headers_disabled(app, users):
    """Test login from headers is disabled."""
    app.login_manager.request_callback = None
    email = users[0]['email']
    basic_fmt = 'Basic {0}'
    decoded = bytes.decode(base64.b64encode(str.encode(str(email))))
    headers = [('Authorization', basic_fmt.format(decoded))]
    with app.test_client() as client:
        res = client.get(
            url_for('invenio_accounts.security'),
            headers=headers,
            environ_base={'REMOTE_ADDR': '127.0.0.1'})
        # check redirect to login
        assert res.status_code == 302
        assert '<a href="/login/' in res.data.decode('utf-8')


def test_disabled_login(app, users):
    """Test login from headers is disabled."""
    app.config["ACCOUNTS_LOCAL_LOGIN_ENABLED"] = False
    email = users[0]['email']
    password = users[0]['password']
    _security.login_form = LoginForm

    with app.test_client() as client:
        url = url_for_security('login')
        data = {'email': email, 'password': password, 'remember': True}

        res = client.post(
            url,
            data=data,
            environ_base={'REMOTE_ADDR': '127.0.0.1'}
        )

        assert res.status_code == 404



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Schemas for user profiles and preferences."""


from flask_babelex import lazy_gettext as _
from marshmallow import Schema, ValidationError, fields


def validate_visibility(value):
    """Check if the value is a valid visibility setting."""
    if value not in ["public", "restricted"]:
        raise ValidationError(
            message=str(_("Value must be either 'public' or 'restricted'."))
        )


class UserProfileSchema(Schema):
    """The default user profile schema."""

    full_name = fields.String()


class UserPreferencesSchema(Schema):
    """The default schema for user preferences."""

    visibility = fields.String(validate=validate_visibility)
    email_visibility = fields.String(validate=validate_visibility)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

import pytest
import requests
from flask import Flask
from flask_babelex import Babel
from flask_mail import Mail
from flask_security import url_for_security
from invenio_db import InvenioDB, db

from invenio_accounts import InvenioAccounts, InvenioAccountsREST, testutils
from invenio_accounts.models import Role, User


def test_version():
    """Test version import."""
    from invenio_accounts import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    app.config['SECRET_KEY'] = 'CHANGEME'
    Babel(app)
    Mail(app)
    InvenioDB(app)
    ext = InvenioAccounts(app)
    assert 'invenio-accounts' in app.extensions
    assert 'security' in app.blueprints.keys()

    app = Flask('testapp')
    app.config['SECRET_KEY'] = 'CHANGEME'
    Babel(app)
    Mail(app)
    InvenioDB(app)
    ext = InvenioAccounts()
    assert 'invenio-accounts' not in app.extensions
    assert 'security' not in app.blueprints.keys()
    ext.init_app(app)
    assert 'invenio-accounts' in app.extensions
    assert 'security' in app.blueprints.keys()


def test_init_rest():
    """Test REST extension initialization."""
    app = Flask('testapp')
    app.config['SECRET_KEY'] = 'CHANGEME'
    Babel(app)
    Mail(app)
    InvenioDB(app)
    ext = InvenioAccountsREST(app)
    assert 'invenio-accounts' in app.extensions
    assert 'security' not in app.blueprints.keys()
    assert 'security_email_templates' in app.blueprints.keys()

    app = Flask('testapp')
    app.config['SECRET_KEY'] = 'CHANGEME'
    Babel(app)
    Mail(app)
    InvenioDB(app)
    ext = InvenioAccountsREST()
    assert 'invenio-accounts' not in app.extensions
    assert 'security' not in app.blueprints.keys()
    ext.init_app(app)
    assert 'invenio-accounts' in app.extensions
    assert 'security' not in app.blueprints.keys()
    assert 'security_email_templates' in app.blueprints.keys()

    app = Flask('testapp')
    app.config['SECRET_KEY'] = 'CHANGEME'
    app.config['ACCOUNTS_REGISTER_BLUEPRINT'] = True
    Babel(app)
    Mail(app)
    InvenioDB(app)
    ext = InvenioAccountsREST()
    assert 'invenio-accounts' not in app.extensions
    assert 'security' not in app.blueprints.keys()
    ext.init_app(app)
    assert 'invenio-accounts' in app.extensions
    assert 'security' in app.blueprints.keys()
    assert 'security_email_templates' in app.blueprints.keys()


def test_alembic(app):
    """Test alembic recipes."""
    ext = app.extensions['invenio-db']

    if db.engine.name == 'sqlite':
        raise pytest.skip('Upgrades are not supported on SQLite.')

    assert not ext.alembic.compare_metadata()
    db.drop_all()
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()
    ext.alembic.downgrade(target='e12419831262')
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()


def test_datastore_usercreate(app):
    """Test create user."""
    ds = app.extensions['invenio-accounts'].datastore

    with app.app_context():
        u1 = ds.create_user(email='info@inveniosoftware.org', password='1234',
                            active=True)
        ds.commit()
        u2 = ds.find_user(email='info@inveniosoftware.org')
        assert u1 == u2
        assert 1 == \
            User.query.filter_by(email='info@inveniosoftware.org').count()


def test_datastore_rolecreate(app):
    """Test create user."""
    ds = app.extensions['invenio-accounts'].datastore

    with app.app_context():
        r1 = ds.create_role(name='superuser', description='1234')
        ds.commit()
        r2 = ds.find_role('superuser')
        assert r1 == r2
        assert 1 == \
            Role.query.filter_by(name='superuser').count()


def test_datastore_assignrole(app):
    """Create and assign user to role."""
    ds = app.extensions['invenio-accounts'].datastore

    with app.app_context():
        u = ds.create_user(email='info@inveniosoftware.org', password='1234',
                           active=True)
        r = ds.create_role(name='superuser', description='1234')
        ds.add_role_to_user(u, r)
        ds.commit()
        u = ds.get_user('info@inveniosoftware.org')
        assert len(u.roles) == 1
        assert u.roles[0].name == 'superuser'


def test_view(app):
    """Test view."""
    with app.app_context():
        login_url = url_for_security('login')

    with app.test_client() as client:
        res = client.get(login_url)
        assert res.status_code == 200


def test_configuration(base_app):
    """Test configuration."""
    app = base_app
    app.config['ACCOUNTS_USE_CELERY'] = 'deadbeef'
    InvenioAccounts(app)
    assert 'deadbeef' == app.config['ACCOUNTS_USE_CELERY']


def test_cookies(cookie_app, users):
    """Test cookies set on login."""
    u = users[0]

    with cookie_app.test_client() as client:
        res = client.post(
            url_for_security('login'),
            data=dict(email=u['email'], password=u['password'], remember=True),
        )
        assert res.status_code == 302
        cookies = {c.name: c for c in client.cookie_jar}
        assert 'session' in cookies
        assert 'remember_token' not in cookies

        # Cookie must be HTTP only, secure and have a domain specified.
        for c in cookies.values():
            assert c.path == '/'
            assert c.domain_specified is True, 'no domain in {}'.format(c.name)
            assert c.has_nonstandard_attr('HttpOnly')
            assert c.secure is True


def test_kvsession_store_init_with_default_factory(app):
    from simplekv.memory import DictStore as kvsession_store_class
    assert isinstance(app.kvsession_store, kvsession_store_class)


def test_kvsession_store_init_with_redis_url(app_with_redis_url):
    from simplekv.memory.redisstore import RedisStore as kvsession_store_class
    assert isinstance(
        app_with_redis_url.kvsession_store, kvsession_store_class)


def test_headers_info(app, users):
    """Test if session and user id is set response header."""
    u = users[0]
    url = url_for_security('change_password')
    with app.app_context():
        with app.test_client() as client:
            response = client.get(url)
            # Not logged in, so only session id available
            assert not testutils.client_authenticated(client)
            assert 'X-Session-ID' in response.headers
            assert 'X-User-ID' not in response.headers
            # Login
            testutils.login_user_via_session(client, email=u['email'])
            response = client.get(url)
            cookie = requests.utils.dict_from_cookiejar(client.cookie_jar)
            assert response.headers['X-Session-ID'] == \
                cookie['session'].split('.')[0]
            assert int(response.headers['X-User-ID']) == u['id']



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Background tasks for accounts."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from flask_mail import Message
from invenio_db import db

from .models import LoginInformation, SessionActivity
from .proxies import current_accounts
from .sessions import delete_session


@shared_task
def send_security_email(data):
    """Celery task to send security email.

    :param data: Contains the email data.
    """
    msg = Message()
    msg.__dict__.update(data)
    current_app.extensions['mail'].send(msg)


@shared_task
def clean_session_table():
    """Automatically clean session table.

    To enable a periodically clean of the session table, you should configure
    the task as a celery periodic task.

    .. code-block:: python

        from datetime import timedelta
        CELERYBEAT_SCHEDULE = {
            'session_cleaner': {
                'task': 'invenio_accounts.tasks.clean_session_table',
                'schedule': timedelta(days=1),
            },
        }

    See `Invenio-Celery <https://invenio-celery.readthedocs.io/>`_
    documentation for further details.
    """
    sessions = SessionActivity.query_by_expired().all()
    for session in sessions:
        delete_session(sid_s=session.sid_s)
    db.session.commit()


@shared_task
def delete_ips():
    """Automatically remove login_info.last_login_ip older than 30 days."""
    expiration_date = datetime.utcnow() - \
        current_app.config['ACCOUNTS_RETENTION_PERIOD']

    LoginInformation.query.filter(
        LoginInformation.last_login_ip.isnot(None),
        LoginInformation.last_login_at < expiration_date
    ).update({
        LoginInformation.last_login_ip: None
    })

    LoginInformation.query.filter(
        LoginInformation.current_login_ip.isnot(None),
        LoginInformation.current_login_at < expiration_date
    ).update({
        LoginInformation.current_login_ip: None
    })
    db.session.commit()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Backend session management for accounts.

The actual backend sessions are provided by ``flask-kvsession``.
This module provides management functionality and populates the
:class:`invenio_accounts.models.SessionActivity` table as new sessions are
created.
"""

from flask import after_this_request, current_app, request, session
from geolite2 import geolite2
from invenio_db import db
from invenio_rest.csrf import reset_token
from ua_parser import user_agent_parser
from werkzeug.local import LocalProxy

from .models import SessionActivity
from .proxies import current_accounts

_sessionstore = LocalProxy(lambda: current_app.kvsession_store)


def _ip2country(ip):
    """Get user country."""
    if ip:
        match = geolite2.reader().get(ip)
        return match.get('country', {}).get('iso_code') if match else None


def _extract_info_from_useragent(user_agent):
    """Extract extra informations from user."""
    parsed_string = user_agent_parser.Parse(user_agent)
    return {
        'os': parsed_string.get('os', {}).get('family'),
        'browser': parsed_string.get('user_agent', {}).get('family'),
        'browser_version': parsed_string.get('user_agent', {}).get('major'),
        'device': parsed_string.get('device', {}).get('family'),
    }


def add_session(session=None):
    r"""Add a session to the SessionActivity table.

    :param session: Flask Session object to add. If None, ``session``
        is used. The object is expected to have a dictionary entry named
        ``"_user_id"`` and a field ``sid_s``
    """
    if '_user_id' not in session and 'user_id' in session:
        # this is to support both flask-login 0.4.x as well as 0.5+,
        # where 'user_id' was renamed to '_user_id'
        session['_user_id'] = session['user_id']

    user_id, sid_s = session['_user_id'], session.sid_s
    with db.session.begin_nested():
        session_activity = SessionActivity(
            user_id=user_id,
            sid_s=sid_s,
            ip=request.remote_addr,
            country=_ip2country(request.remote_addr),
            **_extract_info_from_useragent(
                request.headers.get('User-Agent', '')
            )
        )
        db.session.merge(session_activity)


def login_listener(app, user):
    """Connect to the user_logged_in signal for table population.

    :param app: The Flask application.
    :param user: The :class:`invenio_accounts.models.User` instance.
    """
    @after_this_request
    def add_user_session(response):
        """Regenerate current session and add to the SessionActivity table.

        .. note:: `flask.session.regenerate()` actually calls Flask-KVSession's
            `flask_kvsession.KVSession.regenerate`.
        """
        # Regenerate the session to avoid session fixation vulnerabilities.
        session.regenerate()
        # Save the session first so that the sid_s gets generated.
        app.session_interface.save_session(app, session, response)
        add_session(session)
        current_accounts.datastore.commit()
        return response


def logout_listener(app, user):
    """Connect to the user_logged_out signal.

    :param app: The Flask application.
    :param user: The :class:`invenio_accounts.models.User` instance.
    """
    @after_this_request
    def _commit(response=None):
        if hasattr(session, 'sid_s'):
            delete_session(session.sid_s)
        # Regenerate the session to avoid session fixation vulnerabilities.
        session.regenerate()
        current_accounts.datastore.commit()
        return response


def csrf_token_reset(app, user):
    """Connect to the user_logged_in signal to reset the csrf token.

    :param app: The Flask application.
    :param user: The :class:`invenio_accounts.models.User` instance.
    """
    reset_token()


def delete_session(sid_s):
    """Delete entries in the data- and kvsessionstore with the given sid_s.

    On a successful deletion, the flask-kvsession store returns 1 while the
    sqlalchemy datastore returns None.

    :param sid_s: The session ID.
    :returns: ``1`` if deletion was successful.
    """
    # Remove entries from sessionstore
    _sessionstore.delete(sid_s)
    # Find and remove the corresponding SessionActivity entry
    with db.session.begin_nested():
        SessionActivity.query.filter_by(sid_s=sid_s).delete()
    return 1


def delete_user_sessions(user):
    """Delete all active user sessions.

    :param user: User instance.
    :returns: If ``True`` then the session is successfully deleted.
    """
    with db.session.begin_nested():
        for s in user.active_sessions:
            _sessionstore.delete(s.sid_s)

        SessionActivity.query.filter_by(user=user).delete()

    return True


def default_session_store_factory(app):
    """Session store factory.

    If ``ACCOUNTS_SESSION_REDIS_URL`` is set, it returns a
    :class:`simplekv.memory.redisstore.RedisStore` otherwise
    a :class:`simplekv.memory.DictStore`.
    """
    accounts_session_redis_url = app.config.get('ACCOUNTS_SESSION_REDIS_URL')
    if accounts_session_redis_url:
        import redis
        from simplekv.memory.redisstore import RedisStore
        return RedisStore(redis.StrictRedis.from_url(
            accounts_session_redis_url))
    from simplekv.memory import DictStore
    return DictStore()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from datetime import datetime, timedelta
from time import sleep

from flask import url_for
from flask_login import login_required
from flask_mail import Message
from flask_security import url_for_security

from invenio_accounts.models import SessionActivity, User
from invenio_accounts.tasks import clean_session_table, delete_ips, \
    send_security_email
from invenio_accounts.testutils import create_test_user


def test_send_message_outbox(task_app):
    """Test sending a security message using Task module."""
    with task_app.app_context():
        with task_app.extensions['mail'].record_messages() as outbox:
            msg = Message('Test1',
                          sender='test1@test1.test1',
                          recipients=['test1@test1.test1'])

            send_security_email(msg.__dict__)

            assert len(outbox) == 1
            sent_msg = outbox[0]
            assert sent_msg.subject == 'Test1'
            assert sent_msg.sender == 'test1@test1.test1'
            assert sent_msg.recipients == ['test1@test1.test1']


def test_send_message_through_security(task_app):
    """Test sending a message through security extension."""
    with task_app.app_context():
        with task_app.extensions['mail'].record_messages() as outbox:
            msg = Message('Test1',
                          sender='test1@test1.test1',
                          recipients=['test1@test1.test1'])

            task_app.extensions['security']._send_mail_task(msg)

            assert len(outbox) == 1
            sent_msg = outbox[0]
            assert sent_msg.subject == 'Test1'
            assert sent_msg.sender == 'test1@test1.test1'
            assert sent_msg.recipients == ['test1@test1.test1']


def test_clean_session_table(task_app):
    """Test clean session table."""
    # set session lifetime
    task_app.permanent_session_lifetime = timedelta(seconds=20)

    # protected page
    @task_app.route('/test', methods=['GET'])
    @login_required
    def test():
        return 'test'

    with task_app.test_request_context():
        user1 = create_test_user(email='user1@invenio-software.org')
        user2 = create_test_user(email='user2@invenio-software.org')

        with task_app.test_client() as client:
            client.post(url_for_security('login'), data=dict(
                email=user1.email,
                password=user1.password_plaintext,
            ))
        assert len(SessionActivity.query.all()) == 1
        sleep(15)

        with task_app.test_client() as client:
            client.post(url_for_security('login'), data=dict(
                email=user2.email,
                password=user2.password_plaintext,
            ))
            assert len(SessionActivity.query.all()) == 2
            sleep(10)

            clean_session_table.s().apply()
            assert len(SessionActivity.query.all()) == 1

            protected_url = url_for('test')

            res = client.get(protected_url)
            assert res.status_code == 200

            sleep(15)
            clean_session_table.s().apply()
            assert len(SessionActivity.query.all()) == 0

            res = client.get(protected_url)
            # check if the user is really logout
            assert res.status_code == 302


def test_delete_ips(task_app):
    """Test if ips are deleted after 30 days."""
    last_login_at1 = datetime.utcnow() - \
        task_app.config['ACCOUNTS_RETENTION_PERIOD'] - timedelta(days=1)
    last_login_at2 = datetime.utcnow()

    with task_app.app_context():
        user1 = create_test_user(email='user1@invenio-software.org',
                                 last_login_ip='127.0.0.1',
                                 current_login_ip='127.0.0.1',
                                 last_login_at=last_login_at1,
                                 current_login_at=last_login_at1)

        user2 = create_test_user(email='user2@invenio-software.org',
                                 last_login_ip='127.0.0.1',
                                 current_login_ip='127.0.0.1',
                                 last_login_at=last_login_at2,
                                 current_login_at=last_login_at2)

        user3 = create_test_user(email='user3@invenio-software.org',
                                 last_login_ip='127.0.0.1',
                                 current_login_ip='127.0.0.1',
                                 last_login_at=last_login_at1,
                                 current_login_at=last_login_at2)

        delete_ips()

        user = User.query.filter(User.id == user1.id).one()
        assert user.last_login_ip is None
        assert user.current_login_ip is None

        user = User.query.filter(User.id == user2.id).one()
        assert user.last_login_ip is not None
        assert user.current_login_ip is not None

        user = User.query.filter(User.id == user3.id).one()
        assert user.last_login_ip is None
        assert user.current_login_ip is not None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test test utilities."""

from datetime import datetime

import flask_login
import pytest
from flask_security import url_for_security
from flask_security.utils import hash_password
from invenio_db import db

from invenio_accounts import testutils
from invenio_accounts.errors import JWTDecodeError, JWTExpiredToken
from invenio_accounts.utils import jwt_create_token, jwt_decode_token


def test_client_authenticated(app):
    """Test for testutils.py:client_authenticated(client).

    We want to verify that it doesn't return True when the client isn't
    authenticated/logged in.
    """
    ds = app.extensions['security'].datastore
    email = 'test@example.org'
    password = '123456'

    with app.app_context():
        change_password_url = url_for_security('change_password')
        login_url = url_for_security('login')

        with app.test_client() as client:
            # At this point we should not be authenticated/logged in as a user
            # assert flask_login.current_user.is_anonymous
            assert not testutils.client_authenticated(
                client, test_url=change_password_url)

            # Test HTTP status code of view when not logged in.
            response = client.get(change_password_url)
            assert response.status_code == 302
            assert change_password_url not in response.location
            assert login_url in response.location

            # Once more, following redirects.
            response = client.get(change_password_url, follow_redirects=True)
            assert response.status_code == 200
            assert response.location is None

            # Create a user manually directly in the datastore
            ds.create_user(email=email, password=hash_password(password))
            db.session.commit()

            # Manual login via view
            response = client.post(login_url,
                                   data={'email': email, 'password': password},
                                   environ_base={'REMOTE_ADDR': '127.0.0.1'})

            # Client gets redirected after logging in
            assert response.status_code == 302
            assert testutils.client_authenticated(client)
            assert flask_login.current_user.is_authenticated
            # `is_authenticated` returns True as long as the user object
            # isn't anonymous, i.e. it's an actual user.

            response = client.get(change_password_url)
            assert response.status_code == 200
            response = client.get(change_password_url, follow_redirects=True)
            assert response.status_code == 200


def test_create_test_user(app):
    """Test for testutils.py:create_test_user('test@example.org').

    Demonstrates basic usage and context requirements.
    """
    ds = app.extensions['security'].datastore
    email = 'test@example.org'
    password = '1234'

    with app.app_context():
        user = testutils.create_test_user(email, password)
        assert user.password_plaintext == password
        # Will fail if the app is configured to not "encrypt" the passwords.
        assert user.password != password

        # Verify that user exists in app's datastore
        user_ds = ds.find_user(email=email)
        assert user_ds
        assert user_ds.password == user.password

        with pytest.raises(Exception):
            # Catch-all "Exception" because it's raised by the datastore,
            # and the actual exception type will probably vary depending on
            # which datastore we're running the tests with.
            testutils.create_test_user(email, password)
        # No more tests here b/c the sqlalchemy session crashes when we try to
        # create a user with a duplicate email.


def test_create_test_user_defaults(app):
    """Test the default values for testutils.py:create_test_user."""
    with app.app_context():
        user = testutils.create_test_user('test@example.org')
        with app.test_client() as client:
            testutils.login_user_via_view(client, user.email,
                                          user.password_plaintext)
            assert testutils.client_authenticated(client)


def test_login_user_via_view(app):
    """Test the login-via-view function/hack."""
    email = 'test@example.org'
    password = '1234'

    with app.app_context():
        user = testutils.create_test_user(email, password)
        with app.test_client() as client:
            assert not testutils.client_authenticated(client)
            testutils.login_user_via_view(client, user.email,
                                          user.password_plaintext)
            assert testutils.client_authenticated(client)


def test_login_user_via_session(app):
    """Test the login-via-view function/hack."""
    email = 'test@example.org'
    password = '1234'

    with app.app_context():
        user = testutils.create_test_user(email, password)
        with app.test_client() as client:
            assert not testutils.client_authenticated(client)
            testutils.login_user_via_session(client, email=user.email)
            assert testutils.client_authenticated(client)


def test_jwt_token(app):
    """Test jwt creation."""
    with app.app_context():
        # Extra parameters
        extra = dict(
            defenders=['jessica', 'luke', 'danny', 'matt']
        )
        # Create token normally
        token = jwt_create_token(user_id=22, additional_data=extra)
        decode = jwt_decode_token(token)
        # Decode
        assert 'jessica' in decode.get('defenders')
        assert 22 == decode.get('sub')


def test_jwt_expired_token(app):
    """Test jwt creation."""
    with app.app_context():
        # Extra parameters
        extra = dict(
            exp=datetime(1970, 1, 1),
        )
        # Create token
        token = jwt_create_token(user_id=1, additional_data=extra)
        # Decode
        with pytest.raises(JWTExpiredToken):
            jwt_decode_token(token)
        # Random token
        with pytest.raises(JWTDecodeError):
            jwt_decode_token('Roadster SV')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2019 CERN.
# Copyright (C)      2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API user management and authentication."""

from functools import wraps

from flask import Blueprint, after_this_request, current_app, jsonify
from flask.views import MethodView
from flask_login import login_required
from flask_security import current_user
from flask_security.confirmable import confirm_email_token_status, \
    confirm_user, requires_confirmation
from flask_security.recoverable import reset_password_token_status, \
    update_password
from flask_security.signals import reset_password_instructions_sent
from flask_security.utils import config_value, get_message, login_user, \
    logout_user, send_mail, verify_and_update_password
from flask_security.views import logout
from invenio_db import db
from invenio_rest.errors import FieldError, RESTValidationError
from webargs import ValidationError, fields, validate
from webargs.flaskparser import FlaskParser as FlaskParserBase

from invenio_accounts.models import SessionActivity
from invenio_accounts.sessions import delete_session

from ..proxies import current_datastore, current_security
from ..utils import change_user_password, default_confirmation_link_func, \
    default_reset_password_link_func, obj_or_import_string, register_user


def user_already_authenticated(f):
    """Return user if already authenticated."""
    @wraps(f)
    def wrapper(*args, **kwargs):
        if current_user.is_authenticated:
            return jsonify(default_user_payload(current_user))
        return f(*args, **kwargs)
    return wrapper


def role_to_dict(role):
    """Serialize a new role to dict.

    :param role: a new role to serialize into dict.
    :return: dict from role.
    :rtype: dict.
    """
    return dict(
        id=role.id,
        name=role.name,
        description=role.description,
    )


def create_blueprint(app):
    """Conditionally creates the blueprint."""
    blueprint = Blueprint('invenio_accounts_rest_auth', __name__)

    security_state = app.extensions['security']

    if app.config['ACCOUNTS_REST_AUTH_VIEWS']:
        # Resolve the view classes
        authentication_views = {
            k: obj_or_import_string(v)
            for k, v in app.config.get('ACCOUNTS_REST_AUTH_VIEWS', {}).items()
        }

        blueprint.add_url_rule(
            '/login',
            view_func=authentication_views['login'].as_view('login'))
        blueprint.add_url_rule(
            '/logout',
            view_func=authentication_views['logout'].as_view('logout'))
        blueprint.add_url_rule(
            '/me',
            view_func=authentication_views['user_info'].as_view('user_info'))

        if security_state.registerable:
            blueprint.add_url_rule(
                '/register',
                view_func=authentication_views['register'].as_view('register'))

        if security_state.changeable:
            blueprint.add_url_rule(
                '/change-password',
                view_func=authentication_views['change_password'].as_view(
                    'change_password'))

        if security_state.recoverable:
            blueprint.add_url_rule(
                '/forgot-password',
                view_func=authentication_views['forgot_password'].as_view(
                    'forgot_password'))
            blueprint.add_url_rule(
                '/reset-password',
                view_func=authentication_views['reset_password'].as_view(
                    'reset_password'))

        if security_state.confirmable:
            blueprint.add_url_rule(
                '/send-confirmation-email',
                view_func=authentication_views['send_confirmation'].as_view(
                    'send_confirmation'))

            blueprint.add_url_rule(
                '/confirm-email',
                view_func=authentication_views['confirm_email'].as_view(
                    'confirm_email'))

        if app.config['ACCOUNTS_SESSION_ACTIVITY_ENABLED']:
            blueprint.add_url_rule(
                '/sessions',
                view_func=authentication_views['sessions_list'].as_view(
                    'sessions_list'))

            blueprint.add_url_rule(
                '/sessions/<sid_s>',
                view_func=authentication_views['sessions_item'].as_view(
                    'sessions_item'))
    return blueprint


class FlaskParser(FlaskParserBase):
    """Parser to add FieldError to validation errors."""

    def handle_error(self, error, *args, **kwargs):
        """Handle errors during parsing."""
        if isinstance(error, ValidationError):
            _errors = []
            for field, messages in error.messages.items():
                _errors.extend([FieldError(field, msg) for msg in messages])
            raise RESTValidationError(errors=_errors)
        super(FlaskParser, self).handle_error(error, *args, **kwargs)


webargs_parser = FlaskParser()
use_args = webargs_parser.use_args
use_kwargs = webargs_parser.use_kwargs


#
# Field validators
#
def user_exists(email):
    """Validate that a user exists."""
    if not current_datastore.get_user(email):
        raise ValidationError(get_message('USER_DOES_NOT_EXIST')[0])


def unique_user_email(email):
    """Validate unique user email."""
    if current_datastore.get_user(email) is not None:
        raise ValidationError(
            get_message('EMAIL_ALREADY_ASSOCIATED', email=email)[0])


def default_user_payload(user):
    """Parse user payload."""
    fmt_last_login_at = None
    if user.login_info and user.login_info.last_login_at:
        fmt_last_login_at = user.login_info.last_login_at.isoformat()

    return {
        'id': user.id,
        'email': user.email,
        'confirmed_at':
            user.confirmed_at.isoformat() if user.confirmed_at else None,
        'last_login_at': fmt_last_login_at,
        'roles': [role_to_dict(role) for role in user.roles],
    }


def _abort(message, field=None, status=None):
    if field:
        raise RESTValidationError([FieldError(field, message)])
    raise RESTValidationError(description=message)


def _commit(response=None):
    current_datastore.commit()
    return response


class UserViewMixin(object):
    """Mixin class for get user operations."""

    def get_user(self, email=None, **kwargs):
        """Retrieve a user by the provided arguments."""
        return current_datastore.get_user(email)


class LoginView(MethodView, UserViewMixin):
    """View to login a user."""

    decorators = [user_already_authenticated]

    post_args = {
        'email': fields.Email(required=True, validate=[user_exists]),
        'password': fields.String(required=True)
    }

    def success_response(self, user):
        """Return a successful login response."""
        return jsonify(default_user_payload(user))

    def verify_login(self, user, password=None, **kwargs):
        """Verify the login via password."""
        if not user.password:
            _abort(get_message('PASSWORD_NOT_SET')[0], 'password')
        if not verify_and_update_password(password, user):
            _abort(get_message('INVALID_PASSWORD')[0], 'password')
        if requires_confirmation(user):
            _abort(get_message('CONFIRMATION_REQUIRED')[0])
        if not user.is_active:
            _abort(get_message('DISABLED_ACCOUNT')[0])

    def login_user(self, user):
        """Perform any login actions."""
        return login_user(user)

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Verify and login a user."""
        if not current_app.config.get("ACCOUNTS_LOCAL_LOGIN_ENABLED"):
            _abort(get_message("LOCAL_LOGIN_DISABLED")[0])

        user = self.get_user(**kwargs)
        self.verify_login(user, **kwargs)
        self.login_user(user)
        return self.success_response(user)


class UserInfoView(MethodView):
    """View to fetch info from current user."""

    decorators = [login_required]

    def success_response(self, user):
        """Return a successful user info response."""
        return jsonify(default_user_payload(user))

    def get(self):
        """Return user info."""
        return self.success_response(current_user)


class LogoutView(MethodView):
    """View to logout a user."""

    def logout_user_without_post_redirect(self):
        """Perform any logout actions."""
        if current_user.is_authenticated:
            logout_user()

    def success_response(self):
        """Return a successful logout response."""
        return jsonify({'message': 'User logged out.'})

    def post(self):
        """Logout a user."""
        self.logout_user_without_post_redirect()
        return self.success_response()

    def get(self):
        """Logout user."""
        return logout()


class RegisterView(MethodView):
    """View to register a new user."""

    decorators = [user_already_authenticated]

    post_args = {
        'email': fields.Email(required=True, validate=[unique_user_email]),
        'password': fields.String(
            required=True, validate=[validate.Length(min=6, max=128)])
    }

    def login_user(self, user):
        """Perform any login actions."""
        if not current_security.confirmable or \
                current_security.login_without_confirmation:
            after_this_request(_commit)
            login_user(user)

    def success_response(self, user):
        """Return a successful register response."""
        return jsonify(default_user_payload(user))

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Register a user."""
        if not current_security.registerable:
            _abort(get_message("REGISTRATION_DISABLED")[0])

        user = register_user(**kwargs)
        self.login_user(user)
        return self.success_response(user)


class ForgotPasswordView(MethodView, UserViewMixin):
    """View to get a link to reset the user password."""

    decorators = [user_already_authenticated]

    reset_password_link_func = default_reset_password_link_func

    post_args = {
        'email': fields.Email(required=True, validate=[user_exists]),
    }

    @classmethod
    def send_reset_password_instructions(cls, user):
        """Send email containing instructions to reset password."""
        token, reset_link = cls.reset_password_link_func(user)
        if config_value('SEND_PASSWORD_RESET_EMAIL'):
            send_mail(config_value('EMAIL_SUBJECT_PASSWORD_RESET'), user.email,
                      'reset_instructions', user=user, reset_link=reset_link)
            reset_password_instructions_sent.send(
                current_app._get_current_object(), user=user, token=token)

    def success_response(self, user):
        """Return a response containing reset password instructions."""
        return jsonify({'message': get_message(
            'PASSWORD_RESET_REQUEST', email=user.email)[0]})

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Send reset password instructions."""
        if not current_security.recoverable:
            _abort(get_message("PASSWORD_RECOVERY_DISABLED")[0])

        user = self.get_user(**kwargs)
        self.send_reset_password_instructions(user)
        return self.success_response(user)


class ResetPasswordView(MethodView):
    """View to reset the user password."""

    decorators = [user_already_authenticated]

    post_args = {
        'token': fields.String(required=True),
        'password': fields.String(
            required=True, validate=[validate.Length(min=6, max=128)]),
    }

    def get_user(self, token=None, **kwargs):
        """Retrieve a user by the provided arguments."""
        # Verify the token
        expired, invalid, user = reset_password_token_status(token)
        if invalid:
            _abort(get_message('INVALID_RESET_PASSWORD_TOKEN')[0])
        if expired:
            _abort(get_message(
                'PASSWORD_RESET_EXPIRED',
                email=user.email,
                within=current_security.reset_password_within)[0])
        return user

    def success_response(self, user):
        """Return a successful reset password response."""
        return jsonify({'message': get_message('PASSWORD_RESET')[0]})

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Reset user password."""
        # TODO there doesn't seem to be a `.resettable` or similar?
        if not current_security.recoverable:
            _abort(get_message("PASSWORD_RESET_DISABLED")[0])

        user = self.get_user(**kwargs)
        after_this_request(_commit)
        update_password(user, kwargs['password'])
        login_user(user)
        return self.success_response(user)


class ChangePasswordView(MethodView):
    """View to change the user password."""

    decorators = [login_required]

    post_args = {
        'password': fields.String(
            required=True, validate=[validate.Length(min=6, max=128)]),
        'new_password': fields.String(
            required=True, validate=[validate.Length(min=6, max=128)]),
    }

    def verify_password(self, password=None, new_password=None, **kwargs):
        """Verify password is not invalid."""
        if not verify_and_update_password(password, current_user):
            _abort(get_message('INVALID_PASSWORD')[0], 'password')
        if password == new_password:
            _abort(get_message('PASSWORD_IS_THE_SAME')[0], 'password')

    def change_password(self, new_password=None, **kwargs):
        """Perform any change password actions."""
        after_this_request(_commit)
        change_user_password(user=current_user._get_current_object(),
                             password=new_password)

    def success_response(self):
        """Return a successful change password response."""
        return jsonify({'message': get_message('PASSWORD_CHANGE')[0]})

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Change user password."""
        if not current_security.changeable:
            _abort(get_message("PASSWORD_CHANGE_DISABLED")[0])

        self.verify_password(**kwargs)
        self.change_password(**kwargs)
        return self.success_response()


class SendConfirmationEmailView(MethodView, UserViewMixin):
    """View function which sends confirmation instructions."""

    decorators = [login_required]

    confirmation_link_func = default_confirmation_link_func

    post_args = {
        'email': fields.Email(required=True, validate=[user_exists]),
    }

    def verify(self, user):
        """Verify that user is not confirmed."""
        if user.confirmed_at is not None:
            _abort(get_message('ALREADY_CONFIRMED')[0])

    @classmethod
    def send_confirmation_link(cls, user):
        """Send confirmation link."""
        send_email_enabled = current_security.confirmable and \
            config_value('SEND_REGISTER_EMAIL')
        if send_email_enabled:
            token, confirmation_link = cls.confirmation_link_func(user)
            send_mail(
                config_value('EMAIL_SUBJECT_REGISTER'), user.email,
                'confirmation_instructions', user=user,
                confirmation_link=confirmation_link)
            return token

    def success_response(self, user):
        """Return a successful confirmation email sent response."""
        return jsonify({
            'message': get_message('CONFIRMATION_REQUEST', email=user.email)[0]
        })

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Send confirmation email."""
        user = self.get_user(**kwargs)
        self.verify(user)
        self.send_confirmation_link(user)
        return self.success_response(user)


class ConfirmEmailView(MethodView):
    """View that handles a email confirmation request."""

    post_args = {
        'token': fields.String(required=True),
    }

    def get_user(self, token=None, **kwargs):
        """Retrieve a user by the provided arguments."""
        expired, invalid, user = confirm_email_token_status(token)

        if not user or invalid:
            _abort(get_message('INVALID_CONFIRMATION_TOKEN'))

        already_confirmed = user is not None and user.confirmed_at is not None
        if expired and not already_confirmed:
            _abort(get_message(
                'CONFIRMATION_EXPIRED',
                email=user.email,
                within=current_security.confirm_email_within))
        return user

    @use_kwargs(post_args)
    def post(self, **kwargs):
        """Confirm user email."""
        user = self.get_user(**kwargs)

        if user != current_user:
            logout_user()

        if confirm_user(user):
            after_this_request(_commit)
            return jsonify({'message': get_message('EMAIL_CONFIRMED')[0]})
        else:
            return jsonify({'message': get_message('ALREADY_CONFIRMED')[0]})


class SessionsListView(MethodView):
    """View that returns the list of user sessions."""

    decorators = [login_required]

    def get(self, **kwargs):
        """Return user sessions info."""
        sessions = SessionActivity.query_by_user(
            user_id=current_user.get_id())
        results = [{
            'created': s.created,
            'current': SessionActivity.is_current(s.sid_s),
            'browser': s.browser,
            'browser_version':  s.browser_version,
            'os': s.os,
            'device': s.device,
            'country': s.country} for s in sessions]

        return jsonify({'total': sessions.count(), 'results': results})


class SessionsItemView(MethodView):
    """View for operations related to user session."""

    decorators = [login_required]

    def delete(self, sid_s=None, **kwargs):
        """Revoke the given user session."""
        if SessionActivity.query_by_user(current_user.get_id()) \
                .filter_by(sid_s=sid_s).count() == 1:
            delete_session(sid_s=sid_s)
            db.session.commit()
            message = 'Session {0} successfully removed. {1}.'
            if SessionActivity.is_current(sid_s=sid_s):
                message = message.format(sid_s, "Logged out")
            else:
                message = message.format(sid_s, "Revoked")
            return jsonify({'message': message})
        else:
            return jsonify({
                'message': 'Unable to remove session {0}.'.format(sid_s)}), 400



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Utility function for ACCOUNTS."""

import re
import uuid
from datetime import datetime
from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

from flask import current_app, request, session, url_for
from flask_babelex import lazy_gettext as _
from flask_security import current_user
from flask_security.confirmable import generate_confirmation_token
from flask_security.recoverable import generate_reset_password_token
from flask_security.signals import password_changed, user_registered
from flask_security.utils import config_value as security_config_value
from flask_security.utils import get_security_endpoint_name, hash_password, \
    send_mail
from jwt import DecodeError, ExpiredSignatureError, decode, encode
from werkzeug.routing import BuildError
from werkzeug.utils import import_string

from .errors import JWTDecodeError, JWTExpiredToken
from .proxies import current_datastore, current_security


def jwt_create_token(user_id=None, additional_data=None):
    """Encode the JWT token.

    :param int user_id: Addition of user_id.
    :param dict additional_data: Additional information for the token.
    :returns: The encoded token.
    :rtype: str

    .. note::
        Definition of the JWT claims:

        * exp: ((Expiration Time) expiration time of the JWT.
        * sub: (subject) the principal that is the subject of the JWT.
        * jti: (JWT ID) UID for the JWT.
    """
    # Create an ID
    uid = str(uuid.uuid4())
    # The time in UTC now
    now = datetime.utcnow()
    # Build the token data
    token_data = {
        'exp': now + current_app.config['ACCOUNTS_JWT_EXPIRATION_DELTA'],
        'sub': user_id or current_user.get_id(),
        'jti': uid,
    }
    # Add any additional data to the token
    if additional_data is not None:
        token_data.update(additional_data)

    # Encode the token and send it back
    encoded_token = encode(
        token_data,
        current_app.config['ACCOUNTS_JWT_SECRET_KEY'],
        current_app.config['ACCOUNTS_JWT_ALOGORITHM']
    )

    if not isinstance(encoded_token, str):
        # the token only needs to be decoded if it isn't a string already...
        encoded_token = encoded_token.decode('utf-8')

    return encoded_token


def jwt_decode_token(token):
    """Decode the JWT token.

    :param str token: Additional information for the token.
    :returns: The token data.
    :rtype: dict
    """
    try:
        return decode(
            token,
            current_app.config['ACCOUNTS_JWT_SECRET_KEY'],
            algorithms=[
                current_app.config['ACCOUNTS_JWT_ALOGORITHM']
            ]
        )
    except DecodeError as exc:
        raise JWTDecodeError() from exc
    except ExpiredSignatureError as exc:
        raise JWTExpiredToken() from exc


def set_session_info(app, response, **extra):
    """Add X-Session-ID and X-User-ID to http response."""
    session_id = getattr(session, 'sid_s', None)
    if session_id:
        response.headers['X-Session-ID'] = session_id
    if current_user.is_authenticated:
        response.headers['X-User-ID'] = current_user.get_id()


def obj_or_import_string(value, default=None):
    """Import string or return object.

    :params value: Import path or class object to instantiate.
    :params default: Default object to return if the import fails.
    :returns: The imported object.
    """
    if isinstance(value, str):
        return import_string(value)
    elif value:
        return value
    return default


def _generate_token_url(endpoint, token):
    try:
        url = url_for(endpoint, token=token, _external=True)
    except BuildError:
        # Try to parse URL and build
        scheme, netloc, path, query, fragment = urlsplit(endpoint)
        scheme = scheme or request.scheme
        netloc = netloc or request.host
        assert netloc
        qs = parse_qs(query)
        if '{token}' in path:
            path = path.format(token=token)
        else:
            qs['token'] = token
        query = urlencode(qs)
        url = urlunsplit((scheme, netloc, path, query, fragment))
    return url


def default_reset_password_link_func(user):
    """Return the reset password link that will be sent to a user via email."""
    token = generate_reset_password_token(user)
    endpoint = current_app.config['ACCOUNTS_RESET_PASSWORD_ENDPOINT'] or \
        get_security_endpoint_name('reset_password')
    return token, _generate_token_url(endpoint, token)


def default_confirmation_link_func(user):
    """Return the confirmation link that will be sent to a user via email."""
    token = generate_confirmation_token(user)
    endpoint = current_app.config['ACCOUNTS_CONFIRM_EMAIL_ENDPOINT'] or \
        get_security_endpoint_name('confirm_email')
    return token, _generate_token_url(endpoint, token)


def register_user(_confirmation_link_func=None, **user_data):
    """Register a user."""
    confirmation_link_func = _confirmation_link_func or \
        default_confirmation_link_func
    if user_data.get('password') is not None:
        user_data['password'] = hash_password(user_data['password'])
    user = current_datastore.create_user(**user_data)
    current_datastore.commit()

    token, confirmation_link = None, None
    if current_security.confirmable and user.confirmed_at is None:
        token, confirmation_link = confirmation_link_func(user)

    user_registered.send(
        current_app._get_current_object(), user=user, confirm_token=token)

    if security_config_value('SEND_REGISTER_EMAIL'):
        send_mail(security_config_value('EMAIL_SUBJECT_REGISTER'), user.email,
                  'welcome', user=user, confirmation_link=confirmation_link)

    return user


def change_user_password(_reset_password_link_func=None, **user_data):
    """Change user password."""
    reset_password_link_func = _reset_password_link_func or \
        default_reset_password_link_func
    user = user_data['user']
    user.password = None
    if user_data.get('password') is not None:
        user.password = hash_password(user_data['password'])
    current_datastore.put(user)
    if security_config_value('SEND_PASSWORD_CHANGE_EMAIL'):
        reset_password_link = None
        if current_security.recoverable:
            _, reset_password_link = reset_password_link_func(user)
        subject = security_config_value('EMAIL_SUBJECT_PASSWORD_CHANGE_NOTICE')
        send_mail(subject, user.email, 'change_notice_rest', user=user,
                  reset_password_link=reset_password_link)
    password_changed.send(current_app._get_current_object(),
                          user=user)


def validate_username(username):
    """Validate the username.

    :param username: The username to validate.
    :raises ValueError: If validation fails.
    """
    username_regex = current_app.config["ACCOUNTS_USERNAME_REGEX"]

    if not re.fullmatch(username_regex, username):
        # if validation fails, we raise a ValueError with the configured
        # text explaining the validation rules.
        message = _(current_app.config["ACCOUNTS_USERNAME_RULES_TEXT"])
        raise ValueError(message)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test token duration."""

import time

import pytest
from flask_security import url_for_security
from flask_security.confirmable import confirm_email_token_status, \
    generate_confirmation_token
from flask_security.recoverable import generate_reset_password_token


@pytest.mark.parametrize("sleep,expired", [
    (0, False),
    (4, True),
])
def test_forgot_password_token(app, users, sleep, expired):
    """Test expiration of token for password reset."""
    token = generate_reset_password_token(users[0]['obj'])
    reset_link = url_for_security('reset_password', token=token)

    with app.test_client() as client:
        res = client.get(reset_link, follow_redirects=True)
        time.sleep(sleep)
        if expired:
            app.config['SECURITY_MSG_PASSWORD_RESET_EXPIRED'][0] % {
                'within': app.config['SECURITY_RESET_PASSWORD_WITHIN'],
                'email': users[0]['email']
            } in res.get_data(as_text=True)
        else:
            assert (
                '<button type="submit" class="btn btn-primary btn-lg '
                'btn-block">Reset Password</button>'
            ) in res.get_data(as_text=True)


def test_confirmation_token(app, users):
    """Test expiration of token for email confirmation.

    Test to ensures that the configuration option is respected.
    """
    user = users[0]['obj']
    token = generate_confirmation_token(user)
    # Valid
    expired, invalid, token_user = confirm_email_token_status(token)
    assert expired is False and invalid is False and token_user is user
    # Expired
    time.sleep(4)
    expired, invalid, token_user = confirm_email_token_status(token)
    assert expired is True and invalid is False and token_user is user



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

import pytest
from flask import current_app, session, url_for
from flask_admin import menu
from flask_security.utils import hash_password
from invenio_db import db
from werkzeug.local import LocalProxy

from invenio_accounts import InvenioAccounts
from invenio_accounts.cli import users_create
from invenio_accounts.models import SessionActivity
from invenio_accounts.testutils import login_user_via_view

_datastore = LocalProxy(
    lambda: current_app.extensions['security'].datastore
)


def test_admin(app, admin_view):
    """Test flask-admin interace."""

    # Test activation and deactivation

    with app.app_context():
        # create user and save url for testing
        request_url = url_for("user.action_view")
        kwargs = dict(email="test@test.cern", active=False,
                      password=hash_password('aafaf4as5fa'))
        _datastore.create_user(**kwargs)
        _datastore.commit()
        inserted_id = _datastore.get_user('test@test.cern').id

    with app.test_client() as client:

        res = client.post(
            request_url,
            data={'rowid': inserted_id, 'action': 'activate'},
            follow_redirects=True
        )
        assert res.status_code == 200

        res = client.post(
            request_url,
            data={'rowid': inserted_id, 'action': 'inactivate'},
            follow_redirects=True
        )
        assert res.status_code == 200

        pytest.raises(
            ValueError, client.post, request_url,
            data={'rowid': -42, 'action': 'inactivate'},
            follow_redirects=True
        )
        pytest.raises(
            ValueError, client.post, request_url,
            data={'rowid': -42, 'action': 'activate'},
            follow_redirects=True
        )


def test_admin_createuser(app, admin_view):
    """Test flask-admin user creation"""

    with app.test_client() as client:
        # Test empty mail form

        res = client.post(
            url_for('user.create_view'),
            data={'email': ''},
            follow_redirects=True
        )
        assert b'This field is required.' in res.data

        # Reproduces the workflow described in #154

        res = client.post(
            url_for('user.create_view'),
            data={'email': 'test1@test.cern'},
            follow_redirects=True
        )
        assert _datastore.get_user('test1@test.cern') is not None

        res = client.post(
            url_for('user.create_view'),
            data={'email': 'test2@test.cern', 'active': 'true'},
            follow_redirects=True
        )
        user = _datastore.get_user('test2@test.cern')
        assert user is not None
        assert user.active is True

        res = client.post(
            url_for('user.create_view'),
            data={'email': 'test3@test.cern', 'active': 'false'},
            follow_redirects=True
        )
        user = _datastore.get_user('test3@test.cern')
        assert user is not None
        assert user.active is False

    user_data = dict(email='test4@test.cern', active=False,
                     password=hash_password('123456'))
    _datastore.create_user(**user_data)

    user_data = dict(email='test5@test.cern', active=True,
                     password=hash_password('123456'))
    _datastore.create_user(**user_data)

    user_data = dict(email='test6@test.cern', active=False,
                     password=hash_password('123456'))
    _datastore.create_user(**user_data)
    _datastore.commit()
    assert _datastore.get_user('test4@test.cern') is not None
    user = _datastore.get_user('test5@test.cern')
    assert user is not None
    assert user.active is True
    user = _datastore.get_user('test6@test.cern')
    assert user is not None
    assert user.active is False


def test_admin_sessions(app, admin_view, users):
    """Test flask-admin session."""
    with app.test_request_context():
        index_view_url = url_for('sessionactivity.index_view')
        delete_view_url = url_for('sessionactivity.delete_view')
    with app.test_client() as client:
        res = client.get(index_view_url)
        assert res.status_code == 200

        # simulate login as user 1
        datastore = app.extensions['security'].datastore
        login_user_via_view(client=client, email=users[0]['email'],
                            password=users[0]['password'])
        from flask import session
        sid_s = session.sid_s
        # and try to delete own session sid_s: FAILS
        res = client.post(
            delete_view_url, data={'id': sid_s}, follow_redirects=True)
        assert res.status_code == 200
        sessions = SessionActivity.query.all()
        assert len(sessions) == 1
        assert sessions[0].sid_s == sid_s

    with app.test_client() as client:
        # simulate login as user 2
        login_user_via_view(client=client, email=users[1]['email'],
                            password=users[1]['password'])
        new_sid_s = session.sid_s
        sessions = SessionActivity.query.all()
        assert len(sessions) == 2
        all_sid_s = [session.sid_s for session in sessions]
        assert sorted([sid_s, new_sid_s]) == sorted(all_sid_s)
        # and try to delete a session of another user: WORKS
        res = client.post(
            delete_view_url, data={'id': sid_s},
            follow_redirects=True)
        sessions = SessionActivity.query.all()
        assert len(sessions) == 1
        assert sessions[0].sid_s == new_sid_s



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio-Accounts utility functions for tests and testing purposes.

.. warning:: DO NOT USE IN A PRODUCTION ENVIRONMENT.

Functions within accessing the datastore will throw an error if called outside
of an application context. If pytest-flask is installed you don't have to worry
about this.
"""

import flask
import flask_login
from flask import current_app
from flask_kvsession import SessionID
from flask_security import url_for_security
from flask_security.utils import hash_password
from werkzeug.local import LocalProxy

# "Convenient references" (lifted from flask_security source)
_datastore = LocalProxy(lambda: current_app.extensions['security'].datastore)


def create_test_user(email, password='123456', **kwargs):
    """Create a user in the datastore, bypassing the registration process.

    Accesses the application's datastore. An error is thrown if called from
    outside of an application context.

    Returns the created user model object instance, with the plaintext password
    as `user.password_plaintext`.

    :param email: The user email.
    :param password: The user password. (Default: ``123456``)
    :returns: A :class:`invenio_accounts.models.User` instance.
    """
    assert flask.current_app.testing
    hashed_password = hash_password(password) if password else None
    user = _datastore.create_user(email=email, password=hashed_password,
                                  **kwargs)
    _datastore.commit()
    user.password_plaintext = password
    return user


def login_user_via_session(client, user=None, email=None):
    """Login a user via the session.

    :param client: The CLI test client.
    :param user: The :class:`invenio_accounts.models.User` instance. Optional.
        (Default: ``None``)
    :param email: Load the user by the email. Optional. (Default: ``None``)
    """
    if not user:
        user = _datastore.find_user(email=email)
    with client.session_transaction() as sess:
        sess['_user_id'] = user.get_id()


def login_user_via_view(client, email=None, password=None, user=None,
                        login_url=None):
    r"""Attempt to log the given user in via the 'login' view on the client.

    :param client: client to send the request from.
    :param email: email of user account to log in with.
    :param password: password of user account to log in with.
    :param user: If present, ``user.email`` and ``user.password_plaintext`` \
        take precedence over the `email` and `password` parameters.
    :type user: :class:`invenio_accounts.models.User` (with the addition of \
        a ``password_plaintext`` field)
    :param login_url: URL to post login details to. Defaults to the current \
        application's login URL.

    :returns: The response object from the POST to the login form.
    """
    if user is not None:
        email = user.email
        password = user.password_plaintext
    return client.post(login_url or url_for_security('login'),
                       data={'email': email, 'password': password},
                       environ_base={'REMOTE_ADDR': '127.0.0.1'})
    # If the REMOTE_ADDR isn't set it'll throw out a ValueError as it attempts
    # to update the User model in the database with 'untrackable' as the new
    # `last_login_ip`.


def client_authenticated(client, test_url=None):
    r"""Attempt to access the change password page with the given client.

    :param test_url: URL to attempt to get. Defaults to the current \
            application's "change password" page.
    :returns: True if the client can get the test_url without getting \
        redirected and ``flask_login.current_user`` is not anonymous \
        after requesting the page.
    """
    response = client.get(test_url or url_for_security('change_password'))

    return (response.status_code == 200 and
            not flask_login.current_user.is_anonymous)


def webdriver_authenticated(webdriver, test_url=None):
    """Attempt to get the change password page through the given webdriver.

    Similar to `client_authenticated`, but for selenium webdriver objects.
    """
    save_url = webdriver.current_url

    webdriver.get(test_url or flask.url_for('security.change_password',
                                            _external=True))
    result_url = webdriver.current_url
    webdriver.get(save_url)
    return (flask.url_for('security.login', _external=True) not in result_url)


def unserialize_session(sid_s):
    """Return the unserialized session.

    :param sid_s: The session ID.
    :returns: The unserialized version.
    """
    return SessionID.unserialize(sid_s)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Module tests."""

from click.testing import CliRunner

from invenio_accounts.cli import roles_add, roles_create, roles_remove, \
    users_activate, users_create, users_deactivate


def test_cli_createuser(script_info):
    """Test create user CLI."""
    runner = CliRunner()

    # Missing params
    result = runner.invoke(
        users_create, input='1234\n1234\n', obj=script_info)
    assert result.exit_code != 0

    # Create user with invalid email
    result = runner.invoke(
        users_create,
        ['not-an-email', '--password', '123456'],
        obj=script_info
    )
    assert result.exit_code == 2

    # Create user
    result = runner.invoke(
        users_create,
        ['info@inveniosoftware.org', '--password', '123456'],
        obj=script_info
    )
    assert result.exit_code == 0


def test_cli_createrole(script_info):
    """Test create user CLI."""
    runner = CliRunner()

    # Missing params
    result = runner.invoke(
        roles_create, ['-d', 'Test description'],
        obj=script_info)
    assert result.exit_code != 0

    # Create role
    result = runner.invoke(
        roles_create,
        ['superusers', '-d', 'Test description'],
        obj=script_info)
    assert result.exit_code == 0


def test_cli_addremove_role(script_info):
    """Test add/remove role."""
    runner = CliRunner()

    # Create a user and a role
    result = runner.invoke(
        users_create,
        ['a@example.org', '--password', '123456'],
        obj=script_info
    )
    assert result.exit_code == 0
    result = runner.invoke(roles_create, ['superuser'], obj=script_info)
    assert result.exit_code == 0

    # User not found
    result = runner.invoke(
        roles_add, ['inval@example.org', 'superuser'],
        obj=script_info)
    assert result.exit_code != 0

    # Add:
    result = runner.invoke(
        roles_add, ['a@example.org', 'invalid'],
        obj=script_info)
    assert result.exit_code != 0

    result = runner.invoke(
        roles_remove, ['inval@example.org', 'superuser'],
        obj=script_info)
    assert result.exit_code != 0

    # Remove:
    result = runner.invoke(
        roles_remove, ['a@example.org', 'invalid'],
        obj=script_info)
    assert result.exit_code != 0

    result = runner.invoke(
        roles_remove, ['b@example.org', 'superuser'],
        obj=script_info)
    assert result.exit_code != 0

    result = runner.invoke(
        roles_remove, ['a@example.org', 'superuser'],
        obj=script_info)
    assert result.exit_code != 0

    # Add:
    result = runner.invoke(roles_add,
                           ['a@example.org', 'superuser'],
                           obj=script_info)
    assert result.exit_code == 0
    result = runner.invoke(
        roles_add,
        ['a@example.org', 'superuser'],
        obj=script_info)
    assert result.exit_code != 0

    # Remove:
    result = runner.invoke(
        roles_remove, ['a@example.org', 'superuser'],
        obj=script_info)
    assert result.exit_code == 0


def test_cli_activate_deactivate(script_info):
    """Test create user CLI."""
    runner = CliRunner()

    # Create a user
    result = runner.invoke(
        users_create,
        ['a@example.org', '--password', '123456'],
        obj=script_info
    )
    assert result.exit_code == 0

    # Activate
    result = runner.invoke(users_activate, ['in@valid.org'],
                           obj=script_info)
    assert result.exit_code != 0
    result = runner.invoke(users_deactivate, ['in@valid.org'],
                           obj=script_info)
    assert result.exit_code != 0

    result = runner.invoke(users_activate, ['a@example.org'],
                           obj=script_info)
    assert result.exit_code == 0
    result = runner.invoke(users_activate, ['a@example.org'],
                           obj=script_info)
    assert result.exit_code == 0

    # Deactivate
    result = runner.invoke(users_deactivate,
                           ['a@example.org'], obj=script_info)
    assert result.exit_code == 0
    result = runner.invoke(users_deactivate,
                           ['a@example.org'], obj=script_info)
    assert result.exit_code == 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Validated dictionary classes for user profiles and preferences."""

from inspect import isclass

from invenio_base.utils import load_or_import_from_config
from marshmallow import ValidationError


class ValidatedDict(dict):
    """A dictionary class that validates itself when properties are set."""

    def __init__(self, schema, *args, **kwargs):
        """Constructor, validates the given data."""
        self._schema = schema() if isclass(schema) else schema
        self._validate(dict(*args, **kwargs))
        super().__init__(*args, **kwargs)

    def _validate(self, data):
        """Validate the data with the dictionary's schema."""
        try:
            if self._schema is not None:
                # without schema, we basically revert to a normal dictionary
                # with more overhead
                self._schema.load(data)
        except ValidationError as error:
            raise ValueError(f"Validation failed: {error}")

    def _try_op(self, op_name, *args, **kwargs):
        """Try the named operation and see if it violates the schema."""
        data = {**self}
        getattr(data, op_name)(*args, **kwargs)
        self._validate(data)

    def clear(self):
        """Remove all items from the dictionary."""
        self._validate({})
        return super().clear()

    def pop(self, key, default=None):
        """Remove specified key and return the corresponding value."""
        self._try_op("pop", key, default)
        return super().pop(key, default)

    def popitem(self, *args, **kwargs):
        """Remove and return a (key, value) pair as a 2-tuple."""
        self._try_op("popitem", *args, **kwargs)
        return super().popitem(*args, **kwargs)

    def update(self, e, **f):
        """Update the dict from dict/iterable e and f."""
        self._try_op("update", e, **f)
        return super().update(e, **f)

    def setdefault(self, key, default=None):
        """Insert key with a value of default if key is not in the dict."""
        self._try_op("setdefault", key, default)
        return super().setdefault(key, default)

    def __setitem__(self, key, value):
        """Validate the dictionary and set the value if successful."""
        data = {**self, key: value}
        self._validate(data)
        super().__setitem__(key, value)

    def __delitem__(self, key):
        """Validate the dictionary and delete the key if successful."""
        data = {**self}
        del data[key]
        self._validate(data)
        super().__delitem__(key)


class UserProfileDict(ValidatedDict):
    """Dictionary that validates itself against the user profile schema."""

    def __init__(self, **kwargs):
        """Constructor."""
        schema = load_or_import_from_config("ACCOUNTS_USER_PROFILE_SCHEMA")
        super().__init__(schema, **kwargs)


class UserPreferenceDict(ValidatedDict):
    """Dictionary that validates itself against the preference schema."""

    def __init__(self, **kwargs):
        """Constructor."""
        schema = load_or_import_from_config("ACCOUNTS_USER_PREFERENCES_SCHEMA")
        super().__init__(schema, **kwargs)



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add user profile and preferences as JSON fields to the User table."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy_utils.types import JSONType

# revision identifiers, used by Alembic.
revision = "eb9743315a9d"
down_revision = "62efc52773d4"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.add_column(
        "accounts_user",
        sa.Column(
            "profile",
            sa.JSON()
            .with_variant(JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()),
                "postgresql",
            )
            .with_variant(JSONType(), "sqlite"),
            nullable=True,
        ),
    )
    op.add_column(
        "accounts_user",
        sa.Column(
            "preferences",
            sa.JSON()
            .with_variant(JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()),
                "postgresql",
            )
            .with_variant(JSONType(), "sqlite"),
            nullable=True,
        ),
    )

    # the user name is split into two columns:
    # 'displayname' which stores the original version of the username, and
    # 'username' which stores a lower-case version to ensure uniqueness
    op.add_column(
        "accounts_user",
        sa.Column("displayname", sa.String(length=255), nullable=True),
    )
    op.add_column(
        "accounts_user",
        sa.Column("username", sa.String(length=255), nullable=True),
    )
    op.create_unique_constraint(
        op.f("uq_accounts_user_username"), "accounts_user", ["username"]
    )


def downgrade():
    """Downgrade database."""
    op.drop_column("accounts_user", "preferences")
    op.drop_column("accounts_user", "profile")
    op.drop_constraint(
        op.f("uq_accounts_user_username"), "accounts_user", type_="unique"
    )
    op.drop_column("accounts_user", "displayname")
    op.drop_column("accounts_user", "username")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Tests for template context processors."""

from flask import render_template_string


def test_context_processor_jwt(app):
    """Test context processor JWT."""
    template = r"""
    {{ jwt() }}
    """
    with app.test_request_context():
        html = render_template_string(template)
        assert 'authorized_token' in html


def test_context_processor_jwt_token(app):
    """Test context processor token JWT."""
    template = r"""
    {{ jwt_token() }}
    """
    with app.test_request_context():
        html = render_template_string(template)
        assert html



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C)      2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio user management and authentication."""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    "pytest-invenio>=1.4.2"
]

extras_require = {
    'admin': [
        'invenio-admin>=1.2.1',
    ],
    'docs': [
        'Sphinx==4.2.0',
    ],
    'mysql': [
        'invenio-db[versioning,mysql]>=1.0.11',
    ],
    'postgresql': [
        'invenio-db[versioning,postgresql]>=1.0.11',
    ],
    'sqlite': [
        'invenio-db[versioning]>=1.0.11',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for name, reqs in extras_require.items():
    if name[0] == ':':
        continue
    if name in ('mysql', 'postgresql', 'sqlite'):
        continue
    extras_require['all'].extend(reqs)

setup_requires = [
    'Babel>=2.8',
]

install_requires = [
    'cryptography>=3.0.0',
    'Flask-Breadcrumbs>=0.4.0',
    'Flask-KVSession-Invenio>=0.6.3',
    'Flask-Mail>=0.9.1',
    'Flask-Menu>=0.5.0',
    'Flask-Security-Invenio>=3.1.1',
    'Jinja2>=3.1.0,<4.0.0',
    'invenio-base>=1.2.5',
    'invenio-celery>=1.2.3',
    'invenio-i18n>=1.3.1',
    'invenio-rest>=1.2.4',
    'invenio-theme>=1.3.4',
    'maxminddb-geolite2>=2017.404',
    'passlib>=1.7.1',
    'pyjwt>=1.5.0',
    'redis>=2.10.5',
    'simplekv>=0.11.2',
    'ua-parser>=0.7.3',
]

packages = find_packages()

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('invenio_accounts', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='invenio-accounts',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='invenio accounts user role login',
    license='MIT',
    author='CERN',
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/invenio-accounts',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
        'flask.commands': [
            'roles = invenio_accounts.cli:roles',
            'users = invenio_accounts.cli:users',
        ],
        'invenio_admin.views': [
            'invenio_accounts_user = invenio_accounts.admin:user_adminview',
            'invenio_accounts_role = invenio_accounts.admin:role_adminview',
            'invenio_accounts_session = '
            'invenio_accounts.admin:session_adminview',
            'invenio_accounts_user_identity = '
            'invenio_accounts.admin:user_identity_adminview',
        ],
        'invenio_base.api_apps': [
            'invenio_accounts_rest = invenio_accounts:InvenioAccountsREST',
        ],
        'invenio_base.apps': [
            'invenio_accounts_ui = invenio_accounts:InvenioAccountsUI',
        ],
        'invenio_base.blueprints': [
            'invenio_accounts = invenio_accounts.views.settings:blueprint',
        ],
        'invenio_base.api_blueprints': [
            'invenio_accounts_rest_auth = '
            'invenio_accounts.views.rest:create_blueprint',
        ],
        'invenio_celery.tasks': [
            'invenio_accounts = invenio_accounts.tasks',
        ],
        'invenio_db.alembic': [
            'invenio_accounts = invenio_accounts:alembic',
        ],
        'invenio_db.models': [
            'invenio_accounts = invenio_accounts.models',
        ],
        'invenio_i18n.translations': [
            'messages = invenio_accounts',
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Development Status :: 5 - Production/Stable',
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio user management and authentication."""

from flask import abort, current_app, flash, redirect, render_template, \
    request, url_for
from flask_babelex import lazy_gettext as _
from flask_breadcrumbs import register_breadcrumb
from flask_login import login_required
from flask_menu import register_menu
from flask_security import current_user
from invenio_db import db
from invenio_theme.proxies import current_theme_icons
from speaklater import make_lazy_string

from ..forms import RevokeForm
from ..models import SessionActivity
from ..sessions import delete_session
from .settings import blueprint


@login_required
@register_menu(
    blueprint, 'settings.security',
    # NOTE: Menu item text (icon replaced by a user icon).
    _('%(icon)s Security', icon=make_lazy_string(
        lambda: f'<i class="{current_theme_icons.shield}"></i>')),
    order=2)
@register_breadcrumb(blueprint, 'breadcrumbs.settings.security', _('Security'))
def security():
    """View for security page."""
    sessions = SessionActivity.query_by_user(
        user_id=current_user.get_id()
    ).all()
    master_session = None
    for index, session in enumerate(sessions):
        if SessionActivity.is_current(session.sid_s):
            master_session = session
            del sessions[index]
    return render_template(
        current_app.config['ACCOUNTS_SETTINGS_SECURITY_TEMPLATE'],
        formclass=RevokeForm,
        sessions=[master_session] + sessions,
        is_current=SessionActivity.is_current
    )


@login_required
def revoke_session():
    """Revoke a session."""
    form = RevokeForm(request.form)
    if not form.validate_on_submit():
        abort(403)

    sid_s = form.data['sid_s']
    if SessionActivity.query.filter_by(
            user_id=current_user.get_id(), sid_s=sid_s).count() == 1:
        delete_session(sid_s=sid_s)
        db.session.commit()
        if not SessionActivity.is_current(sid_s=sid_s):
            # if it's the same session doesn't show the message, otherwise
            # the session will be still open without the database record
            flash('Session {0} successfully removed.'.format(sid_s), 'success')
    else:
        flash('Unable to remove the session {0}.'.format(sid_s), 'error')
    return redirect(url_for('invenio_accounts.security'))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C)      2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio-accounts views."""

from flask import abort, current_app, request
from flask_security.views import anonymous_user_required
from flask_security.views import login as base_login

from .settings import blueprint


@anonymous_user_required
@blueprint.route("/login")
def login(*args, **kwargs):
    """Disable login credential submission if local login is disabled."""
    local_login_enabled = current_app.config.get(
        "ACCOUNTS_LOCAL_LOGIN_ENABLED", True
    )

    login_form_submitted = (request.method == "POST")
    if login_form_submitted and not local_login_enabled:
        # only allow GET requests,
        # avoid credential submission/login via POST
        abort(404)

    return base_login(*args, **kwargs)


__all__ = ('blueprint', 'login')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for the user profile utilities."""

import pytest
from marshmallow import Schema, fields

from invenio_accounts.profiles.dicts import ValidatedDict


class RequiredFieldSchema(Schema):
    """Schema with one optional and one required field."""

    optional = fields.String(required=False)
    required = fields.String(required=True)


class OptionalFieldsSchema(Schema):
    """Schema with only one optional field."""

    optional = fields.String()


def test_validated_dict_required_fields():
    """Test if the removal of required fields fails."""
    # we can leave out the optional field
    dictionary = ValidatedDict(RequiredFieldSchema, required="ok")
    assert "required" in dictionary
    assert "optional" not in dictionary

    # removal of the required field shouldn't work...
    with pytest.raises(ValueError):
        dictionary.pop("required", None)

    with pytest.raises(ValueError):
        del dictionary["required"]

    with pytest.raises(ValueError):
        dictionary.clear()

    # we can't leave out the required field
    with pytest.raises(ValueError):
        ValidatedDict(RequiredFieldSchema, optional="ok")


def test_validated_dict_optional_fields():
    """Test if we can delete optional fields."""
    # test a few ways of deleting optional fields
    dictionary = ValidatedDict(OptionalFieldsSchema, optional="ok")
    assert dictionary
    dictionary.pop("optional", None)
    assert not dictionary

    dictionary = ValidatedDict(OptionalFieldsSchema, optional="ok")
    del dictionary["optional"]
    assert not dictionary

    dictionary = ValidatedDict(OptionalFieldsSchema, optional="ok")
    dictionary.clear()
    assert not dictionary

    dictionary = ValidatedDict(OptionalFieldsSchema, optional="ok")


def test_validated_dict():
    """A few general tests for the validated dictionary."""
    # unexpected values should raise an error
    with pytest.raises(ValueError):
        ValidatedDict(OptionalFieldsSchema, required="ok")

    # setting wrong types should raise an error
    with pytest.raises(ValueError):
        ValidatedDict(OptionalFieldsSchema, optional=123)

    # update should work...
    dictionary = ValidatedDict(OptionalFieldsSchema)
    assert "optional" not in dictionary
    dictionary.update({"optional": "yes"})
    assert dictionary["optional"] == "yes"

    # ... but not with a wrong type
    with pytest.raises(ValueError):
        dictionary.update({"optional": False})

    # setdefault should work as expected
    dictionary = ValidatedDict(RequiredFieldSchema, required="ok")
    assert "optional" not in dictionary
    dictionary.setdefault("optional", "yes")
    assert dictionary["optional"] == "yes"
    dictionary.setdefault("optional", "no")
    assert dictionary["optional"] == "yes"

    # setting values per key should also work
    dictionary["required"] = "absolutely"
    assert dictionary["required"] == "absolutely"
    with pytest.raises(ValueError):
        dictionary["required"] = None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C)      2022 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test invenio-accounts models."""

import pytest
from invenio_db import db
from marshmallow import Schema, fields
from sqlalchemy import inspect

from invenio_accounts import testutils
from invenio_accounts.models import SessionActivity, User


class CustomProfile(Schema):
    """A custom user profile schema."""

    file_descriptor = fields.Integer(strict=True)


def test_session_activity_model(app):
    """Test SessionActivity model."""
    with app.app_context():
        # SessionActivity table is in the database
        inspector = inspect(db.engine)
        assert 'accounts_user_session_activity' in inspector.get_table_names()

        user = testutils.create_test_user('test@example.org')

        # Create a new SessionActivity object, put it in the db
        session_activity = SessionActivity(user_id=user.get_id(),
                                           sid_s="teststring")
        database = db

        # the `created` field is magicked in via the Timestamp mixin class
        assert not session_activity.created
        database.session.add(session_activity)
        # Commit it to the books.
        database.session.commit()
        assert session_activity.created
        assert len(user.active_sessions) == 1

        # Now how does this look on the user object?
        assert session_activity == user.active_sessions[0]

        session_two = SessionActivity(user_id=user.get_id(),
                                      sid_s="testring_2")
        database.session.add(session_two)
        # Commit it to the books.
        database.session.commit()

        assert len(user.active_sessions) == 2
        # Check #columns in table
        queried = database.session.query(SessionActivity)
        assert queried.count() == 2
        active_sessions = queried.all()
        assert session_activity.sid_s in [x.sid_s for x in active_sessions]
        assert session_two in queried.filter(
            SessionActivity.sid_s == session_two.sid_s)
        assert queried.count() == 2  # `.filter` doesn't change the query

        # Test session deletion
        session_to_delete = user.active_sessions[0]
        database.session.delete(session_to_delete)
        assert len(user.active_sessions) == 2  # Not yet updated.
        assert queried.count() == 1
        # Deletion is visible on `user` once database session is commited.
        database.session.commit()
        assert len(user.active_sessions) == 1
        assert user.active_sessions[0].sid_s != session_to_delete.sid_s


def test_profiles(app):
    """Test the user profile."""
    user = User(email="admin@inveniosoftware.org")
    profile = {
        "full_name": "Invenio Admin",
    }

    with pytest.raises(ValueError):
        # the profile doesn't expect an 'email' value
        user.profile = {
            **profile, "email": "admin@inveniosoftware.org",
        }

    assert user.profile is None

    # a valid profile should be accepted
    user.profile = profile
    assert dict(user.profile) == profile

    # setting expected properties should work
    assert len(user.profile) == 1
    assert user.profile["full_name"] == "Invenio Admin"

    # but setting unexpected properties should not work
    with pytest.raises(ValueError):
        user.profile["invalid"] = "value"

    # similar with wrong types for expected fields
    with pytest.raises(ValueError):
        user.profile["email"] = 1

    assert len(user.profile) == 1
    assert user.profile["full_name"] == "Invenio Admin"


def test_custom_profiles(app):
    """Test if the customization mechanism for user profiles works."""
    app.config["ACCOUNTS_USER_PROFILE_SCHEMA"] = CustomProfile()
    user = User(email="admin@inveniosoftware.org")

    # the default fields aren't allowed in the custom schema
    with pytest.raises(ValueError):
        user.profile = {
            "full_name": "Invenio Admin",
        }

    # the expected properties should work...
    user.profile = {"file_descriptor": 1}
    assert dict(user.profile) == {"file_descriptor": 1}

    # ... but not with unexpected types!
    with pytest.raises(ValueError):
        user.profile["file_descriptor"] = "1"

    assert dict(user.profile) == {"file_descriptor": 1}
